/*
    Copyright 2013 Albert Vaca <albertvaka@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) version 3, or any
    later version accepted by the membership of KDE e.V. (or its
    successor approved by the membership of KDE e.V.), which shall
    act as a proxy defined in Section 6 of version 3 of the license.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library. If not, see <http://www.gnu.org/licenses/>.
*/

#include "phonepowermanager.h"
#include "phonepowerdevice.h"

#include <QtDBus/QDBusReply>
#include <QtCore/QDebug>
#include <QtDBus/QDBusMetaType>
#include <QtDBus/QDBusConnectionInterface>
#include <QDBusServiceWatcher>

#include "../shared/rootdevice.h"

using namespace Solid::Backends::PhonePower;
using namespace Solid::Backends::Shared;

PhonePowerManager::PhonePowerManager(QObject *parent)
    : Solid::Ifaces::DeviceManager(parent)
    , m_manager("org.kde.kdeconnect",
                "/modules/kdeconnect",
                QDBusConnection::sessionBus())
    , m_ready(false)
{
    m_supportedInterfaces << Solid::DeviceInterface::Battery;


    QDBusServiceWatcher* watcher = new QDBusServiceWatcher("org.kde.kdeconnect", QDBusConnection::sessionBus(), QDBusServiceWatcher::WatchForOwnerChange, this);
    connect(watcher, SIGNAL(serviceRegistered(QString)), this, SLOT(serviceReady()));
    connect(watcher, SIGNAL(serviceUnregistered(QString)), this, SLOT(serviceUnregistered()));

    QDBusPendingCall msg = m_manager.asyncCall("visibleDevices");
    msg.waitForFinished();
    if (msg.isValid()) {
        serviceReady();
    }

}

void PhonePowerManager::serviceReady()
{

    QDBusPendingReply<QStringList> pendingDeviceList = m_manager.visibleDevices();
    pendingDeviceList.waitForFinished();
    if (!pendingDeviceList.isValid() || pendingDeviceList.isError()) {
        return;
    }
    QStringList deviceList = pendingDeviceList.value();

    Q_FOREACH(const QString& device, deviceList) {
        slotDeviceAdded(device);
    }

    connect(&m_manager, SIGNAL(deviceVisibilityChanged(QString, bool)),
            this, SLOT(deviceVisibilityChanged(QString,bool)));

    m_ready = true;
}

void PhonePowerManager::serviceUnregistered()
{
    m_ready = false;
    Q_FOREACH(const QString& device, m_devices) {
        slotDeviceRemoved(device);
    }
    m_devices.clear();

    disconnect(&m_manager, SIGNAL(deviceVisibilityChanged(QString, bool)),
            this, SLOT(deviceVisibilityChanged(QString,bool)));

}

void PhonePowerManager::deviceVisibilityChanged(const QString& id, bool visible)
{
    if (visible)
        slotDeviceAdded(id);
    else
        slotDeviceRemoved(id);
}

PhonePowerManager::~PhonePowerManager()
{

}

QObject* PhonePowerManager::createDevice(const QString& udi)
{
    if (!m_ready) return 0;

    QString connectId = udi.mid(udiPrefix().count() + 1);
    qDebug() << "createDevice" << udi << connectId;
    if (udi == udiPrefix()) {

        RootDevice *root = new RootDevice(udiPrefix());

        root->setProduct(tr("Power Management"));
        root->setDescription(tr("Batteries and other sources of power"));
        root->setIcon("preferences-system-power-management");

        return root;

    } else if (m_devices.contains(connectId)) {
        return new PhonePowerDevice(udi, connectId);
    } else {
        return 0;
    }
}

QStringList PhonePowerManager::devicesFromQuery(const QString& parentUdi, Solid::DeviceInterface::Type type)
{
    if (!m_ready) {
        return QStringList();
    }

    if (type != Solid::DeviceInterface::Battery)
    {
        return QStringList();
    }

    if (!parentUdi.isEmpty() && parentUdi != udiPrefix())
    {
        return QStringList();
    }

    return allDevices();

}

QStringList PhonePowerManager::allDevices()
{
    if (!m_ready) {
        return QStringList();
    }

    QStringList udiList;
    Q_FOREACH(const QString& device, m_devices) {
        udiList.append(udiPrefix()+"/"+device);
    }

    return udiList;
}

QSet< Solid::DeviceInterface::Type > PhonePowerManager::supportedInterfaces() const
{
    return m_supportedInterfaces;
}

QString PhonePowerManager::udiPrefix() const
{
    return QLatin1String("/org/kde/kdeconnect");
}

void PhonePowerManager::slotDeviceAdded(const QString &id)
{

    qDebug() << "slotDeviceAdded" << id;

    if (id.isEmpty() || m_devices.contains(id)) {
        return;
    }

    //Test if device has battery support!
    QDBusInterface* device = new QDBusInterface("org.kde.kdeconnect", "/modules/kdeconnect/devices/"+id, "org.kde.kdeconnect.device", QDBusConnection::sessionBus(), this);
    QDBusPendingReply<bool> reply = device->asyncCallWithArgumentList("hasPlugin",QVariantList() << QVariant("kdeconnect_battery"));
    reply.waitForFinished();
    if (!reply.isValid()) {
        return;
    }

    bool hasBattery = reply.value();

    qDebug() << "HasBattery:" << hasBattery;

    if (hasBattery) {
        m_devices.insert(id);
        Q_EMIT deviceAdded(udiPrefix()+"/"+id);
        device->deleteLater();
    } else {
        connect(device, SIGNAL(pluginsChanged()),
                this, SLOT(pluginsChanged()));
    }

}

void PhonePowerManager::pluginsChanged()
{
    QDBusInterface* device = (QDBusInterface*)sender();

    disconnect(device, SIGNAL(pluginsChanged()),
               this, SLOT(pluginsChanged()));

    QVariant id = device->property("id");
    if (id.isValid()) {
        qDebug() << "PluginsChanged" << id.toString();
        QDBusInterface* device = new QDBusInterface("org.kde.kdeconnect", "/modules/kdeconnect/devices/"+id.toString(), "org.kde.kdeconnect.device", QDBusConnection::sessionBus(), this);
        if (!device->isValid()) return;
        QDBusPendingReply<bool> reply = device->asyncCallWithArgumentList("hasPlugin",QVariantList() << QVariant("kdeconnect_battery"));
        reply.waitForFinished();
        if (!reply.isValid() || !reply.value()) {
            slotDeviceRemoved(id.toString());
        } else {
            slotDeviceAdded(id.toString());
        }

        //device->deleteLater();
    }

}

void PhonePowerManager::slotDeviceRemoved(const QString &id)
{
    qDebug() << "slotDeviceRemoved" << id;
    if (!m_devices.contains(id)) {
        return;
    }

    m_devices.remove(id);
    Q_EMIT deviceRemoved(id);
}

#include "backends/phonepower/phonepowermanager.moc"
